/*
 *  Copyright 2009-2010 Intel Corporation.  All Rights Reserved.
 *
 *  The source code contained or described herein and all documents related
 *  to the source code ("Material") are owned by Intel Corporation or its
 *  suppliers or licensors.  Title to the Material remains with Intel
 *  Corporation or its suppliers and licensors.  The Material is protected
 *  by worldwide copyright laws and treaty provisions.  No part of the
 *  Material may be used, copied, reproduced, modified, published, uploaded,
 *  posted, transmitted, distributed, or disclosed in any way without
 *  Intel's prior express written permission.
 *
 *  No license under any patent, copyright, trade secret or other
 *  intellectual property right is granted to or conferred upon you by
 *  disclosure or delivery of the Materials, either expressly, by
 *  implication, inducement, estoppel or otherwise.  Any license under such
 *  intellectual property rights must be express and approved by Intel in
 *  writing.
 *
 */

/*
 * reducer_min.h
 *
 * Purpose: Reducer hyperobject to retain the min value.
 *
 * Classes: reducer_min<Type>
 *          reducer_min_index<Index, Value>
 *
 * Description:
 * ============
 * This component provides reducer-type hyperobject representations that allow
 * the minimum value, or the minimum value and an index, of a group of values to
 * be determined in parallel.
 *
 * Usage Example:
 * ==============
 * Suppose we wish to compute the minimum value in an array of integers.
 *..
 *  int test()
 *  {
 *      int a[ARRAY_SIZE];
 *      int min = INT_MIN;
 *
 *      ...
 *
 *      for (int i = 0; i < ARRAY_SIZE; ++i)
 *      {
 *          if (a[i] < min)
 *          {
 *              min = a[i];
 *          }
 *      }
 *      std::cout << "min = " << min << std::endl;
 *
 *      ...
 *  }
 *..
 * Changing the 'for' to a 'cilk_for' will allow the loop to be run in parallel
 * but will create a data race on the variable 'min'.  The race can be resolved
 * by changing 'min' to a 'reducer_min' hyperobject:
 *..
 *  int test()
 *  {
 *      int a[ARRAY_SIZE];
 *      cilk::reducer_min<int> min(INT_MIN);
 *
 *      ...
 *
 *      cilk_for (int i = 0; i < ARRAY_SIZE; ++i)
 *      {
 *          min = cilk::min_of(min, a[i]);
 *      }
 *      std::cout << "min = " << min.get_value() << std::endl;
 *
 *      ...
 *  }
 *..
 * A similar loop which calculates both the minimum value and index would be:
 *..
 *  int test()
 *  {
 *      int a[ARRAY_SIZE];
 *      cilk::reducer_min_index<int, int> rmi(INT_MIN, -1);
 *
 *      ...
 *
 *      cilk_for (int i = 0; i < ARRAY_SIZE; ++i)
 *      {
 *          rmi.min_of(i, a[i]);
 *      }
 *      std::cout << "min = " << rmi.get_value() <<
 *                   ", index = " << rmi.get_index() << std::endl;
 *
 *      ...
 *  }
 *..
 *
 * Operations provided:
 * ====================
 * reducer_min and reducer_min_index provide set and get methods that are
 * guaranteed to be deterministic iff they are called prior to the first
 * spawn or after the last sync in a parallel algorithm.  When called during
 * execution, the value returned by get_value (and get_index) may differ from
 * run to run depending on how the routine or loop is scheduled.  Calling
 * set_value anywhere between the first spawn and the last sync may cause the
 * algorithm to produce non-deterministic results.
 *..
 * get_value and get_index return imutable values.  The matching get_reference
 * and get_index_reference methods return modifiable references
 *..
 * The calc_min method is a comparison operation that sets the reducer to the
 * lower of itself and the object being compared.  The min_of routines are
 * provided for convenience:
 *..
 *  cilk::reducer_min<int> rm;
 *
 *  ...
 *
 *  rm = min_of(rm, 55); // alternatively: rm.calc_min(55);
 *..
 *
 * Template parameter restrictions:
 * ================================
 * reducer_min and reducer_min_index require that the 'Type' template parameter
 * implement operator > for use by calc_min.  operator > must implement a total
 * ordering if you want determanistic results.  However, a strict weak ordering
 * is good enough for many algorithms.
 *..
 * There are no requirements on the 'Index' template parameter of
 * reducer_min_index.  All comparisons will be done on the 'Type' value.
 *..
 */

#ifndef REDUCER_MIN_H_INCLUDED
#define REDUCER_MIN_H_INCLUDED

#include <cilk/reducer.h>
#ifdef __cplusplus
# include <cstddef>
#else
# include <stddef.h>
#endif

#ifdef __cplusplus

/* C++ Interface
 */

#include <cassert>

namespace cilk {

// Class 'reducer_min' is a hyperobject representation of a value that
// retains the minimum value of all of the values it sees during its
// lifetime.
template <typename Type>
class reducer_min
{
public:
    typedef Type basic_value_type;

public:
    struct View
    {
        friend class reducer_min<Type>;
        friend class monoid_base<View>;

    public:
        View();
        void take_min(const Type &v);
        void set(const Type &v);

        const Type &get_value() const;
        bool is_set() const;

    private:
        Type value_;
        bool isSet_;
    };

public:
    // View of reducer data
    struct Monoid: monoid_base<View>
    {
        static void reduce (View *left, View *right);
    };
private:
    // Hyperobject to serve up views
    reducer<Monoid> imp_;

public:
    // "PRIVATE" HELPER CLASS - uses the type system to make sure that
    // reducer_min instances aren't copied, but we can still allow statements
    // like *min = cilk::min_of(*min, a[i]);
    class temp_min
    {
    private:
        friend class reducer_min;

        // Copyable, not assignable
        temp_min& operator=(const temp_min &);

        explicit temp_min(reducer_min *reducerPtr);

    private:
        reducer_min* reducerPtr_;

    public:
        temp_min calc_min(const Type& x) const;
    };

    friend class temp_min;

public:
    // Construct a 'reducer_min' object with a value of 'Type()'.
    reducer_min();

    // Construct a 'reducer_min' object with the specified initial value.
    explicit reducer_min(const Type& initial_value);

    // Return an immutable reference to the value of this object.
    const Type& get_value() const;

    // Return a reference to the value of this object.
    Type&       get_reference();
    Type const& get_reference() const;

    // Returns true if the value has ever been set
    bool is_set() const;

    // Set the value of this object.  Warning: Setting the value of a
    // hyperobject such that it violates the associative operation
    // algebra will yield results that are likely to differ from serial
    // execution and may differ from run to run.
    void set_value(const Type& value);

    // Compare the current value with the one passed and retain the
    // larger of the two.
    temp_min calc_min(const Type& value);

    // Merge the result of a 'min' operation into this object.  The
    // operation must involve this hyperobject, i.e., x = min_of(x, 5);
    reducer_min& operator=(const temp_min &temp);

private:
    // Not copyable
    reducer_min(const reducer_min&);
    reducer_min& operator=(const reducer_min&);
};

// Global "cilk::min_of" functions

template <typename Type>
inline
typename reducer_min<Type>::temp_min
min_of(reducer_min<Type>& r,
       const Type& x)
{
    return r.calc_min(x);
}

template <typename Type>
inline
typename reducer_min<Type>::temp_min
min_of(const typename reducer_min<Type>::temp_min& tmp,
       const Type& x)
{
    return tmp.calc_min(x);
}

template <typename Type>
inline
typename reducer_min<Type>::temp_min
min_of(const Type& x,
       reducer_min<Type>& r)
{
    return r.calc_min(x);
}

template <typename Type>
inline
typename reducer_min<Type>::temp_min
min_of(const Type& x,
       const typename reducer_min<Type>::temp_min& tmp)
{
    return tmp.calc_min(x);
}

/////////////////////////////////////////////////////////////////////////////
// Implementation of inline and template functions
/////////////////////////////////////////////////////////////////////////////

// --------------------------------
// template class reducer_min::View
// --------------------------------

template<typename Type>
reducer_min<Type>::View::View() :
    value_(),
    isSet_(false)
{
}

template<typename Type>
void reducer_min<Type>::View::set(const Type &v)
{
    value_ = v;
    isSet_ = true;
}

template<typename Type>
void reducer_min<Type>::View::take_min(const Type &v)
{
    if (! isSet_ || (value_ > v))
    {
        value_ = v;
        isSet_ = true;
    }
}

template<typename Type>
const Type &reducer_min<Type>::View::get_value() const
{
    return value_;
}

template<typename Type>
bool reducer_min<Type>::View::is_set() const
{
    return isSet_;
}

// -------------------------------------------
// template class reducer_min::Monoid
// -------------------------------------------

// reduce - Minimises "right" value into "left"
template<typename Type>
void
reducer_min<Type>::Monoid::reduce(View *left, View *right)
{
    // If the right instance hasn't been initialized, there's no need to
    // go any further
    if (! right->isSet_)
        return;

    // Take the min of the two values
    left->take_min (right->value_);
}

// -----------------------------------
// temp_min private helper class implementation
// -----------------------------------

template <typename Type> inline
reducer_min<Type>::temp_min::temp_min(reducer_min<Type> *reducerPtr) :
    reducerPtr_(reducerPtr)
{
}

template <typename Type> inline
typename reducer_min<Type>::temp_min
reducer_min<Type>::temp_min::calc_min(const Type& x) const
{
    return reducerPtr_->calc_min(x);
}

// --------------------------
// template class reducer_min
// --------------------------

// Default constructor
template <typename Type>
inline
reducer_min<Type>::reducer_min()
    : imp_()
{
}

template <typename Type>
inline
reducer_min<Type>::reducer_min(const Type& initial_value)
    : imp_()
{
    View &v = imp_.view();

    v.set(initial_value);
}

template <typename Type>
inline
const Type& reducer_min<Type>::get_value() const
{
    const View &v = imp_.view();

    return v.value_;
}

template <typename Type>
inline
Type& reducer_min<Type>::get_reference()
{
    View &v = imp_.view();

    return v.value_;
}

template <typename Type>
inline
Type const& reducer_min<Type>::get_reference() const
{
    View &v = imp_.view();

    return v.value_;
}

template <typename Type>
inline
bool reducer_min<Type>::is_set() const
{
    const View &v = imp_.view();

    return v.isSet_;
}

template <typename Type>
inline
void reducer_min<Type>::set_value(const Type& value)
{
    View &v = imp_.view();

    v.set(value);
}

template <typename Type> inline
typename reducer_min<Type>::temp_min
reducer_min<Type>::calc_min(const Type& value)
{
    View &v = imp_.view();

    // If no previous value has been set, always set the value

    v.take_min(value);

    return temp_min(this);
}

template <typename Type>
reducer_min<Type>&
reducer_min<Type>::operator=(const typename reducer_min<Type>::temp_min &temp)
{
    // Noop.  Just test that temp is the same as this.
    assert(this == temp.reducerPtr_);
    return *this;
}


// Class 'reducer_min_index' is a hyperobject representation of an index
// and corresponding value representing the minimum such pair this object
// has seen.
template <typename Index, typename Value>
class reducer_min_index
{
public:
    typedef Value basic_value_type;

public:
    struct View
    {
        friend class reducer_min_index<Index, Value>;
        friend class monoid_base<View>;

    public:
        View();
        void set(const Index &i, const Value &v);
        void take_min(const Index &i, const Value &v);

        const Index &get_index() const;
        const Value &get_value() const;
        bool is_set() const;

    private:
        Index index_;
        Value value_;
        bool isSet_;
    };

public:
    // View of reducer data
    struct Monoid: monoid_base<View>
    {
        static void reduce (View *left, View *right);
    };

private:
    // Hyperobject to serve up views
    reducer<Monoid> imp_;

public:
    // Construct a 'reducer_min_index' object with a value of 'Type()'.
    reducer_min_index();


    // Construct a 'reducer_min_index' object with the specified initial
    // value and index.
    explicit reducer_min_index(const Index& initial_index,
                               const Value& initial_value);

    // Return an imutable reference to the value of this object.
    const Value& get_value() const;

    // Return a reference to the value of this object
    Value&       get_reference();
    Value const& get_reference() const;

    // Return an imutable reference to the minimum index.
    const Index& get_index() const;

    // Return a mutable reference to the minimum index
    Index& get_index_reference();

    // Returns true if the value has ever been set
    bool is_set() const;

    // Set the index/value of this object.  Warning: Setting the value of
    // a hyperobject such that it violates the associative operation
    // algebra will yield results that are likely to differ from serial
    // execution and may differ from run to run.
    void set_value(const Index& index,
                   const Value& value);

    // Compare the current value with the one passed and retain the
    // larger of the two.
    void min_of(const Index& index, const Value& value);

private:
    // Not copyable
    reducer_min_index(const reducer_min_index&);
    reducer_min_index& operator=(const reducer_min_index&);
};

/////////////////////////////////////////////////////////////////////////////
// Implementation of inline and template functions
/////////////////////////////////////////////////////////////////////////////

// --------------------------------
// template class reducer_min::View
// --------------------------------

template<typename Index, typename Value>
reducer_min_index<Index, Value>::View::View() :
    index_(),
    value_(),
    isSet_(false)
{
}

template<typename Index, typename Value>
void
reducer_min_index<Index, Value>::View::set(const Index &i, const Value &v)
{
    index_ = i;
    value_ = v;
    isSet_ = true;
}

template<typename Index, typename Value>
void
reducer_min_index<Index, Value>::View::take_min(const Index &i, const Value &v)
{
    if (! isSet_ || (value_ > v))
    {
        index_ = i;
        value_ = v;
        isSet_ = true;
    }
}

template<typename Index, typename Value>
const Index &
reducer_min_index<Index, Value>::View::get_index() const
{
    return index_;
}

template<typename Index, typename Value>
const Value &
reducer_min_index<Index, Value>::View::get_value() const
{
    return value_;
}

template<typename Index, typename Value>
bool
reducer_min_index<Index, Value>::View::is_set() const
{
    return isSet_;
}

// -------------------------------------------
// template class reducer_min::Monoid
// -------------------------------------------

// reduce - Minimises "right" value into "left"
template<typename Index, typename Value>
void
reducer_min_index<Index, Value>::Monoid::reduce(View *left, View *right)
{
    // If the right instance hasn't been initialized, there's no need to
    // go any further
    if (! right->isSet_)
        return;

    left->take_min (right->index_, right->value_);
}

// --------------------------------
// template class reducer_min_index
// --------------------------------

// Default constructor
template <typename Index, typename Value>
inline
reducer_min_index<Index, Value>::reducer_min_index()
    : imp_()
{
}

template <typename Index, typename Value>
inline
reducer_min_index<Index, Value>::reducer_min_index(const Index& initial_index,
                                                   const Value& initial_value)
    : imp_()
{
    View &v = imp_.view();

    v.set(initial_index, initial_value);
}

template <typename Index, typename Value>
inline
void reducer_min_index<Index, Value>::min_of(const Index& index,
                                             const Value& value)
{
    View &v = imp_.view();

    v.take_min(index, value);
}

template <typename Index, typename Value>
inline
const Value& reducer_min_index<Index, Value>::get_value() const
{
    const View &v = imp_.view();

    return v.value_;
}

template <typename Index, typename Value>
inline
Value& reducer_min_index<Index, Value>::get_reference()
{
    View &v = imp_.view();

    return v.value_;
}

template <typename Index, typename Value>
inline
Value const& reducer_min_index<Index, Value>::get_reference() const
{
    const View &v = imp_.view();

    return v.value_;
}

template <typename Index, typename Value>
inline
const Index& reducer_min_index<Index, Value>::get_index() const
{
    const View &v = imp_.view();

    return v.index_;
}

template <typename Index, typename Value>
inline
Index& reducer_min_index<Index, Value>::get_index_reference()
{
    View &v = imp_.view();

    return v.index_;
}

template <typename Index, typename Value>
inline
bool reducer_min_index<Index, Value>::is_set() const
{
    const View &v = imp_.view();

    return v.isSet_;
}

template <typename Index, typename Value>
inline
void reducer_min_index<Index, Value>::set_value(const Index& index,
                                                const Value& value)
{
    View &v = imp_.view();

    return v.set(index, value);
}

} // namespace cilk

#endif // __cplusplus

/* C Interface
 */

__CILKRTS_BEGIN_EXTERN_C

/* REDUCER_MIN */

#define CILK_C_REDUCER_MIN_TYPE(tn)                                         \
    __CILKRTS_MKIDENT(cilk_c_reducer_min_,tn)
#define CILK_C_REDUCER_MIN(obj,tn,v)                                        \
    CILK_C_REDUCER_MIN_TYPE(tn) obj =                                       \
        CILK_C_INIT_REDUCER(_Typeof(obj.value),                             \
                        __CILKRTS_MKIDENT(cilk_c_reducer_min_reduce_,tn),   \
                        __CILKRTS_MKIDENT(cilk_c_reducer_min_identity_,tn), \
                        __cilkrts_hyperobject_noop_destroy, v)

/* Declare an instance of the reducer for a specific numeric type */
#define CILK_C_REDUCER_MIN_INSTANCE(t,tn)                                \
    typedef CILK_C_DECLARE_REDUCER(t)                                    \
        __CILKRTS_MKIDENT(cilk_c_reducer_min_,tn);                       \
    __CILKRTS_DECLARE_REDUCER_REDUCE(cilk_c_reducer_min,tn,l,r);         \
    __CILKRTS_DECLARE_REDUCER_IDENTITY(cilk_c_reducer_min,tn);  

/* CILK_C_REDUCER_MIN_CALC(reducer, v) performs the reducer lookup
 * AND calc_min operation, leaving the current view with the min of the
 * previous value and v.
 */
#define CILK_C_REDUCER_MIN_CALC(reducer, v) do {                            \
    _Typeof((reducer).value)* view = &(REDUCER_VIEW(reducer));              \
    _Typeof(v) __value = (v);                                               \
    if (*view > __value) {                                                  \
        *view = __value;                                                    \
    } } while (0)

/* Declare an instance of the reducer type for each numeric type */
CILK_C_REDUCER_MIN_INSTANCE(char,char);
CILK_C_REDUCER_MIN_INSTANCE(unsigned char,uchar);
CILK_C_REDUCER_MIN_INSTANCE(signed char,schar);
CILK_C_REDUCER_MIN_INSTANCE(wchar_t,wchar_t);
CILK_C_REDUCER_MIN_INSTANCE(short,short);
CILK_C_REDUCER_MIN_INSTANCE(unsigned short,ushort);
CILK_C_REDUCER_MIN_INSTANCE(int,int);
CILK_C_REDUCER_MIN_INSTANCE(unsigned int,uint);
CILK_C_REDUCER_MIN_INSTANCE(unsigned int,unsigned); /* alternate name */
CILK_C_REDUCER_MIN_INSTANCE(long,long);
CILK_C_REDUCER_MIN_INSTANCE(unsigned long,ulong);
CILK_C_REDUCER_MIN_INSTANCE(long long,longlong);
CILK_C_REDUCER_MIN_INSTANCE(unsigned long long,ulonglong);
CILK_C_REDUCER_MIN_INSTANCE(float,float);
CILK_C_REDUCER_MIN_INSTANCE(double,double);
CILK_C_REDUCER_MIN_INSTANCE(long double,longdouble);

/* Declare function bodies for the reducer for a specific numeric type */
#define CILK_C_REDUCER_MIN_IMP(t,tn,id)                                  \
    __CILKRTS_DECLARE_REDUCER_REDUCE(cilk_c_reducer_min,tn,l,r)          \
        { if (*(t*)l > *(t*)r) *(t*)l = *(t*)r; }                        \
    __CILKRTS_DECLARE_REDUCER_IDENTITY(cilk_c_reducer_min,tn)            \
        { *(t*)v = id; }

/* c_reducers.c contains definitions for all of the monoid functions
   for the C numeric tyeps.  The contents of reducer_min.c are as follows:

CILK_C_REDUCER_MIN_IMP(char,char,CHAR_MAX)
CILK_C_REDUCER_MIN_IMP(unsigned char,uchar,CHAR_MIN)
CILK_C_REDUCER_MIN_IMP(signed char,schar,SCHAR_MAX)
CILK_C_REDUCER_MIN_IMP(wchar_t,wchar_t,WCHAR_MAX)
CILK_C_REDUCER_MIN_IMP(short,short,SHRT_MAX)
CILK_C_REDUCER_MIN_IMP(unsigned short,ushort,USHRT_MAX)
CILK_C_REDUCER_MIN_IMP(int,int,INT_MAX)
CILK_C_REDUCER_MIN_IMP(unsigned int,uint,UINT_MAX)
CILK_C_REDUCER_MIN_IMP(unsigned int,unsigned,UINT_MAX) // alternate name
CILK_C_REDUCER_MIN_IMP(long,long,LONG_MAX)
CILK_C_REDUCER_MIN_IMP(unsigned long,ulong,ULONG_MAX)
CILK_C_REDUCER_MIN_IMP(long long,longlong,LLONG_MAX)
CILK_C_REDUCER_MIN_IMP(unsigned long long,ulonglong,ULLONG_MAX)
CILK_C_REDUCER_MIN_IMP(float,float,HUGE_VALF)
CILK_C_REDUCER_MIN_IMP(double,double,HUGE_VAL)
CILK_C_REDUCER_MIN_IMP(long double,longdouble,HUGE_VALL)

*/

/* REDUCER_MIN_INDEX */

#define CILK_C_REDUCER_MIN_INDEX_VIEW(t,tn)                                  \
    typedef struct {                                                         \
        __STDNS ptrdiff_t index;                                             \
        t                 value;                                             \
    } __CILKRTS_MKIDENT(cilk_c_reducer_min_index_view_,tn)

#define CILK_C_REDUCER_MIN_INDEX_TYPE(t) \
    __CILKRTS_MKIDENT(cilk_c_reducer_min_index_,t)
#define CILK_C_REDUCER_MIN_INDEX(obj,t,v)                                    \
    CILK_C_REDUCER_MIN_INDEX_TYPE(t) obj =                                   \
    CILK_C_INIT_REDUCER(_Typeof(obj.value),                                  \
        __CILKRTS_MKIDENT(cilk_c_reducer_min_index_reduce_,t),               \
        __CILKRTS_MKIDENT(cilk_c_reducer_min_index_identity_,t),             \
        __cilkrts_hyperobject_noop_destroy, { 0, v })

/* Declare an instance of the reducer for a specific numeric type */
#define CILK_C_REDUCER_MIN_INDEX_INSTANCE(t,tn)                                  \
    CILK_C_REDUCER_MIN_INDEX_VIEW(t,tn);                                         \
    typedef CILK_C_DECLARE_REDUCER(                                              \
        __CILKRTS_MKIDENT(cilk_c_reducer_min_index_view_,tn))                    \
            __CILKRTS_MKIDENT(cilk_c_reducer_min_index_,tn);                     \
    __CILKRTS_DECLARE_REDUCER_REDUCE(cilk_c_reducer_min_index,tn,l,r);           \
    __CILKRTS_DECLARE_REDUCER_IDENTITY(cilk_c_reducer_min_index,tn);  

/* CILK_C_REDUCER_MIN_INDEX_CALC(reducer, i, v) performs the reducer lookup
 * AND calc_min operation, leaving the current view with the min of the
 * previous value and v.
 */
#define CILK_C_REDUCER_MIN_INDEX_CALC(reducer, i, v) do {                   \
    _Typeof((reducer).value)* view = &(REDUCER_VIEW(reducer));              \
    _Typeof(v) __value = (v);                                               \
    if (view->value > __value) {                                            \
        view->index = (i);                                                  \
        view->value = __value;                                              \
    } } while (0)

/* Declare an instance of the reducer type for each numeric type */
CILK_C_REDUCER_MIN_INDEX_INSTANCE(char,char);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(unsigned char,uchar);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(signed char,schar);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(wchar_t,wchar_t);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(short,short);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(unsigned short,ushort);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(int,int);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(unsigned int,uint);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(unsigned int,unsigned); /* alternate name */
CILK_C_REDUCER_MIN_INDEX_INSTANCE(long,long);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(unsigned long,ulong);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(long long,longlong);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(unsigned long long,ulonglong);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(float,float);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(double,double);
CILK_C_REDUCER_MIN_INDEX_INSTANCE(long double,longdouble);

/* Declare function bodies for the reducer for a specific numeric type */
#define CILK_C_REDUCER_MIN_INDEX_IMP(t,tn,id)                                  \
    __CILKRTS_DECLARE_REDUCER_REDUCE(cilk_c_reducer_min_index,tn,l,r)          \
        { typedef __CILKRTS_MKIDENT(cilk_c_reducer_min_index_view_,tn) view_t; \
          if (((view_t*)l)->value > ((view_t*)r)->value)                       \
              *(view_t*)l = *(view_t*)r; }                                     \
    __CILKRTS_DECLARE_REDUCER_IDENTITY(cilk_c_reducer_min_index,tn)            \
        { typedef __CILKRTS_MKIDENT(cilk_c_reducer_min_index_view_,tn) view_t; \
          ((view_t*)v)->index = 0; ((view_t*)v)->value = id; }

/* c_reducers.c contains definitions for all of the monoid functions
   for the C numeric tyeps.  The contents of reducer_min_index.c are as follows:

CILK_C_REDUCER_MIN_INDEX_IMP(char,char,CHAR_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(unsigned char,uchar,CHAR_MIN)
CILK_C_REDUCER_MIN_INDEX_IMP(signed char,schar,SCHAR_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(wchar_t,wchar_t,WCHAR_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(short,short,SHRT_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(unsigned short,ushort,USHRT_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(int,int,INT_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(unsigned int,uint,UINT_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(unsigned int,unsigned,UINT_MAX) // alternate name
CILK_C_REDUCER_MIN_INDEX_IMP(long,long,LONG_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(unsigned long,ulong,ULONG_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(long long,longlong,LLONG_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(unsigned long long,ulonglong,ULLONG_MAX)
CILK_C_REDUCER_MIN_INDEX_IMP(float,float,HUGE_VALF)
CILK_C_REDUCER_MIN_INDEX_IMP(double,double,HUGE_VAL)
CILK_C_REDUCER_MIN_INDEX_IMP(long double,longdouble,HUGE_VALL)

*/


__CILKRTS_END_EXTERN_C

#endif // defined REDUCER_MIN_H_INCLUDED
