/*
 *  Copyright 2009-2010 Intel Corporation.  All Rights Reserved.
 *
 *  The source code contained or described herein and all documents related
 *  to the source code ("Material") are owned by Intel Corporation or its
 *  suppliers or licensors.  Title to the Material remains with Intel
 *  Corporation or its suppliers and licensors.  The Material is protected
 *  by worldwide copyright laws and treaty provisions.  No part of the
 *  Material may be used, copied, reproduced, modified, published, uploaded,
 *  posted, transmitted, distributed, or disclosed in any way without
 *  Intel's prior express written permission.
 *
 *  No license under any patent, copyright, trade secret or other
 *  intellectual property right is granted to or conferred upon you by
 *  disclosure or delivery of the Materials, either expressly, by
 *  implication, inducement, estoppel or otherwise.  Any license under such
 *  intellectual property rights must be express and approved by Intel in
 *  writing.
 *
 */

/*
 * reducer_ostream.h
 *
 * Purpose: Hyper-object to write to 'std::ostream's
 *
 * Classes: reducer_ostream
 *
 * Description:
 * ============
 * Output streams ('std::ostream's) are a convenient means of writing text to
 * files, the user console, or sockets.  In a serial program, text is written
 * to an ostream in a specific, logical order.  For example, computing while
 * traversing a data structure and printing them to an 'ostream' will result
 * in the values being printed in the order of traversal.  In a parallel
 * version of the same program, however, different parts of the data structure
 * may be traversed in a different order, resulting in a non-deterministic
 * ordering of the stream.  Worse, multiple strands may write to the same
 * stream simultaneously, resulting in a data race.  Replacing the
 * 'std::ostream' with a 'cilk::reducer_ostream' will solve both problems: Data
 * will appeaer in the stream in the same order as it would for the serial
 * program, and there will be no races (no locks) on the common stream.
 *
 * Usage Example:
 * ==============
 * Assume we wish to traverse an array of objects, performing an operation on
 * each object and writing the result to a file.  Without a reducer_ostream,
 * we have a race on the 'output' file stream:
 *..
 *  void compute(std::ostream& os, double x)
 *  {
 *      // Perform some significant computation and print the result:
 *      os << std::asin(x);
 *  }
 *
 *  int test()
 *  {
 *      const std::size_t ARRAY_SIZE = 1000000;
 *      extern double myArray[ARRAY_SIZE];
 *
 *      std::ofstream output("output.txt");
 *      cilk_for (std::size_t i = 0; i < ARRAY_SIZE; ++i)
 *      {
 *          compute(output, myArray[i]);
 *      }
 *
 *      return 0;
 *  }
 *..
 * The race is solved by using a reducer_ostream to proxy the 'output' file:
 *..
 *  void compute(cilk::reducer_ostream& os, double x)
 *  {
 *      // Perform some significant computation and print the result:
 *      os << std::asin(x);
 *  }
 *
 *  int test()
 *  {
 *      const std::size_t ARRAY_SIZE = 1000000;
 *      extern double myArray[ARRAY_SIZE];
 *
 *      std::ofstream output("output.txt");
 *      cilk::reducer_ostream hyper_output(output);
 *      cilk_for (std::size_t i = 0; i < ARRAY_SIZE; ++i)
 *      {
 *          compute(hyper_output, myArray[i]);
 *      }
 *
 *      return 0;
 *  }
 *..
 *
 * Limitations:
 * ============
 * There are two possible values for the formatting flags immediately after a
 * 'cilk_spawn' statement: they may either have the value that was set by the
 * spawn function, or they may have default values.  Because of
 * non-determinism in the processor scheduling, there is no way to determine
 * which it will be.  Similarly, the formatting flags after a 'cilk_sync' may
 * or may not have the same value as before the sync.  Therefore, one must use
 * a disciplined coding style to avoid formatting errors.  There are two
 * approaches to mitigating the problem: The first is to eliminate the
 * difference between the two possible outcomes by ensuring that the spawned
 * function always returns the flags to their initial state:
 *..
 *  void compute(cilk::reducer_ostream& os, double x)
 *  {
 *      // Perform some significant computation and print the result:
 *      int saveprec = os.precision(5);
 *      os << std::asin(x);
 *      os.precision(saveprec);
 *  }
 *..
 * The second approach is to write your streaming operations such that they
 * don't depend on the previous state of the formatting flags by setting any
 * important flags before every block of output:
 *..
 *      cilk_spawn compute(hyper_output, value);
 *
 *      hyper_output->precision(2);  // Don't depend on previous precision
 *      *hyper_output << f();
 *      *hyper_output << g();
 *..
 * Another concern is memory usage.  A reducer_ostream will buffer as much text
 * as necessary to ensure that the order of output matches that of the serial
 * version of the program.  If all spawn branches perform an equal amount of
 * output, then one can expect that half of the output before a sync will be
 * buffered in memory.  This hyperobject is therefore not well suited for
 * serializing very large quantities of text output.
 */

#ifndef REDUCER_OSTREAM_H_INCLUDED
#define REDUCER_OSTREAM_H_INCLUDED

#include <cilk/reducer.h>
#include <iostream>
#include <sstream>

namespace cilk {

// Class 'reducer_ostream' is the representation of a hyperobject for output
// text streaming.
class reducer_ostream
{
public:
    class View: public std::ostream
    {
    public:
        typedef std::ostream Base;

        friend class reducer_ostream;

        View():
            std::ostream(0)
        {
            Base::rdbuf(&strbuf_);
        };

    private:
        void use_ostream (const std::ostream &os)
        {
            Base::rdbuf(os.rdbuf());
            Base::flags(os.flags());       // Copy formatting flags
            Base::setstate(os.rdstate());  // Copy error state
        }

    private:
        std::stringbuf  strbuf_;
    };

public:
    // View of reducer data
    struct Monoid: monoid_base< View >
    {
        static void reduce (View *left, View *right);
    };

private:
    // Hyperobject to serve up views
    reducer<Monoid> imp_;

    // Methods that provide the API for the reducer
public:

    // Construct an initial 'reducer_ostream' from an 'std::ostream'.  The
    // specified 'os' stream is used as the eventual destination for all
    // text streamed to this hyperobject.
    explicit reducer_ostream(const std::ostream &os);

    // Return a modifiable reference to the underlying 'ostream' object.
    std::ostream& get_reference();

    template<typename T>
    std::ostream &
    operator<< (const T &v)
    {
        return imp_.view() << v;
    }

    std::ostream &
    operator<< (std::ostream &(*_Pfn)(std::ostream &))
    {
        View &v = imp_.view();

        return ((*_Pfn)(v));
    }
};


// -------------------------------------------
// class reducer_ostream::Monoid
// -------------------------------------------

// reduce - appends string from "right" reducer_basic_string onto the end of
// the "left". When done, the "right" reducer_basic_string is empty.
void
reducer_ostream::Monoid::reduce(View *left, View *right)
{
    left->operator<< (&right->strbuf_);
}

// --------------------------
// class reducer_ostream
// --------------------------

inline
reducer_ostream::reducer_ostream(const std::ostream &os) :
    imp_()
{
    View &v = imp_.view();

    v.use_ostream(os);
}

inline
std::ostream &
reducer_ostream::get_reference()
{
    View &v = imp_.view();

    return v;
}

} // namespace cilk

#endif //  REDUCER_OSTREAM_H_INCLUDED

